/**
 ****************************************************************************************
 *
 * @file app.h
 *
 * @brief Application entry point
 *
 * Copyright (C) RivieraWaves 2009-2015
 *
 *
 ****************************************************************************************
 */

#ifndef APP_H_
#define APP_H_

/**
 ****************************************************************************************
 * @addtogroup APP
 * @ingroup RICOW
 *
 * @brief Application entry point.
 *
 * @{
 ****************************************************************************************
 */

/*
 * INCLUDE FILES
 ****************************************************************************************
 */

#include "rwip_config.h" // SW configuration

#if (BLE_APP_PRESENT)

#include <stdint.h> // Standard Integer Definition
#include <co_bt.h>  // Common BT Definitions
#include "arch.h"   // Platform Definitions
#include "gapc.h"   // GAPC Definitions
#include "nvds.h"
#include "user_config.h"
/*
 * DEFINES
 ****************************************************************************************
 */
/// Maximal length of the Device Name value
#define APP_DEVICE_NAME_MAX_LEN (APP_DEVICE_NAME_LENGTH_MAX) //(20)

/// Default Advertising duration - 30s (in multiple of 10ms)
#define APP_DFLT_ADV_DURATION (3000)

#define APP_UART_BUF_LENGTH              (256)
/*
 * TYPE DEFINITIONS
 ****************************************************************************************
 */

/// List of Application NVDS TAG identifiers
enum app_nvds_tag
{
    /// BLE Application Advertising data
    NVDS_TAG_APP_BLE_ADV_DATA = 0x0B,
    NVDS_LEN_APP_BLE_ADV_DATA = 32,

    /// BLE Application Scan response data
    NVDS_TAG_APP_BLE_SCAN_RESP_DATA = 0x0C,
    NVDS_LEN_APP_BLE_SCAN_RESP_DATA = 32,

    /// Mouse Sample Rate
    NVDS_TAG_MOUSE_SAMPLE_RATE = 0x38,
    NVDS_LEN_MOUSE_SAMPLE_RATE = 1,
    /// Peripheral Bonded
    NVDS_TAG_PERIPH_BONDED = 0x39,
    NVDS_LEN_PERIPH_BONDED = 1,
    /// Mouse NTF Cfg
    NVDS_TAG_MOUSE_NTF_CFG = 0x3A,
    NVDS_LEN_MOUSE_NTF_CFG = 2,
    /// Mouse Timeout value
    NVDS_TAG_MOUSE_TIMEOUT = 0x3B,
    NVDS_LEN_MOUSE_TIMEOUT = 2,
    /// Peer Device BD Address
    NVDS_TAG_PEER_BD_ADDRESS = 0x3C,
    NVDS_LEN_PEER_BD_ADDRESS = 7,
    /// Mouse Energy Safe
    NVDS_TAG_MOUSE_ENERGY_SAFE = 0x3D,
    NVDS_LEN_MOUSE_SAFE_ENERGY = 2,
    /// EDIV (2bytes), RAND NB (8bytes),  LTK (16 bytes), Key Size (1 byte)
    NVDS_TAG_LTK = 0x3E,
    NVDS_LEN_LTK = 28,
    /// PAIRING
    NVDS_TAG_PAIRING = 0x3F,
    NVDS_LEN_PAIRING = 54,
};

enum app_loc_nvds_tag
{
    /// Audio mode 0 task
    NVDS_TAG_AM0_FIRST = NVDS_TAG_APP_SPECIFIC_FIRST,     // 0x90
    NVDS_TAG_AM0_LAST = NVDS_TAG_APP_SPECIFIC_FIRST + 16, // 0xa0

    /// Local device Identity resolving key
    NVDS_TAG_LOC_IRK,
    /// Peer device Resolving identity key (+identity address)
    NVDS_TAG_PEER_IRK,

    /// size of local identity resolving key
    NVDS_LEN_LOC_IRK = KEY_LEN,
    /// size of Peer device identity resolving key (+identity address)
    NVDS_LEN_PEER_IRK = sizeof(struct gapc_irk),
};

/// Application environment structure
struct app_env_tag
{
    /// Connection handle
    uint16_t conhdl;
    /// Connection Index
    uint8_t conidx;

    /// Last initialized profile
    uint8_t next_svc;

    /// Bonding status
    bool bonded;

    /// Device Name length
    uint8_t dev_name_len;
    /// Device Name
    uint8_t dev_name[APP_DEVICE_NAME_MAX_LEN];

    /// Local device IRK
    uint8_t loc_irk[KEY_LEN];
};

/*
 * GLOBAL VARIABLE DECLARATION
 ****************************************************************************************
 */

/// Application environment
extern struct app_env_tag app_env;

/*
 * FUNCTION DECLARATIONS
 ****************************************************************************************
 */

/**
 ****************************************************************************************
 * @brief Initialize the BLE demo application.
 ****************************************************************************************
 */
void appm_init(void);

/**
 ****************************************************************************************
 * @brief Add a required service in the database
 ****************************************************************************************
 */
bool appm_add_svc(void);

/**
 ****************************************************************************************
 * @brief Put the device in general discoverable and connectable mode
 ****************************************************************************************
 */
void appm_start_advertising(void);

/**
 ****************************************************************************************
 * @brief Put the device in non discoverable and non connectable mode
 ****************************************************************************************
 */
void appm_stop_advertising(void);

/**
 ****************************************************************************************
 * @brief Send to request to update the connection parameters
 ****************************************************************************************
 */
void appm_update_param(struct gapc_conn_param *conn_param);

/**
 ****************************************************************************************
 * @brief Send a disconnection request
 ****************************************************************************************
 */
void appm_disconnect(void);

/**
 ****************************************************************************************
 * @brief Retrieve device name
 *
 * @param[out] device name
 *
 * @return name length
 ****************************************************************************************
 */
uint8_t appm_get_dev_name(uint8_t *name);
void app_user_entry(void);
void app_user_init(void);

// user define
enum at_cmd_tag
{
    AT_RESULT_AT, // 0
    AT_RESULT_BAUD1,
    AT_RESULT_BAUD2,
    AT_RESULT_BAUD3,
    AT_RESULT_BAUD4,
    AT_RESULT_BAUD5,
    AT_RESULT_BAUDS1, // 6
    AT_RESULT_BAUDS2,
    AT_RESULT_BAUDS3,
    AT_RESULT_BAUDS4,
    AT_RESULT_BAUDS5,
    AT_RESULT_NAME,
    AT_RESULT_NAMES,
    AT_RESULT_SUUID,
    AT_RESULT_SUUIDS,
    AT_RESULT_WUUID,
    AT_RESULT_WUUIDS,
    AT_RESULT_NUUID,
    AT_RESULT_NUUIDS,
    AT_RESULT_ADVI1,
    AT_RESULT_ADVI2,
    AT_RESULT_ADVI3,
    AT_RESULT_ADVI4,
    AT_RESULT_ADVI5,
    AT_RESULT_ADVIS1,
    AT_RESULT_ADVIS2,
    AT_RESULT_ADVIS3,
    AT_RESULT_ADVIS4,
    AT_RESULT_ADVIS5,
    AT_RESULT_RFLASH,
    AT_RESULT_WFLASH,
    AT_RESULT_TPMODE0,
    AT_RESULT_TPMODE1,
    AT_RESULT_MAX,
};
enum at_rsp_tag
{
    AT_RESULT_OK0,
    AT_RESULT_OK1,
    AT_RESULT_OK2,
    AT_RESULT_OK3,
    AT_RESULT_OK4,
    AT_RESULT_OK5,
    AT_RESULT_ADVOK1,
    AT_RESULT_ADVOK2,
    AT_RESULT_ADVOK3,
    AT_RESULT_ADVOK4,
    AT_RESULT_ADVOK5,
    AT_RESULT_OKDATA,
    AT_RESULT_OKMODE0,
    AT_RESULT_OKMODE1,
};

#define BACT_STATE_ACTIVE 1
#define BACT_STATE_DEEPSLEEP 0
#define DECLARE_PREFIX(str, code)  \
    {                              \
        str, sizeof(str) - 1, code \
    }
typedef struct
{
    const char *prefix; /* Command prefix, e.g. AT+BRSF= */
    uint8_t prefix_len;
    uint8_t code; /* AT_COMMAND_... or AT_RESULT_... */
} at_prefix_t;

static const at_prefix_t at_prefix_req[] = {
    DECLARE_PREFIX("AT\r\n", AT_RESULT_AT),

    DECLARE_PREFIX("AT+BAUDS1\r\n", AT_RESULT_BAUDS1),
    DECLARE_PREFIX("AT+BAUDS2\r\n", AT_RESULT_BAUDS2),
    DECLARE_PREFIX("AT+BAUDS3\r\n", AT_RESULT_BAUDS3),
    DECLARE_PREFIX("AT+BAUDS4\r\n", AT_RESULT_BAUDS4),
    DECLARE_PREFIX("AT+BAUDS5\r\n", AT_RESULT_BAUDS5),
    DECLARE_PREFIX("AT+BAUD1\r\n", AT_RESULT_BAUD1),
    DECLARE_PREFIX("AT+BAUD2\r\n", AT_RESULT_BAUD2),
    DECLARE_PREFIX("AT+BAUD3\r\n", AT_RESULT_BAUD3),
    DECLARE_PREFIX("AT+BAUD4\r\n", AT_RESULT_BAUD4),
    DECLARE_PREFIX("AT+BAUD5\r\n", AT_RESULT_BAUD5),

    DECLARE_PREFIX("AT+NAMES", AT_RESULT_NAMES),
    DECLARE_PREFIX("AT+NAME", AT_RESULT_NAME),

    DECLARE_PREFIX("AT+SUUIDS", AT_RESULT_SUUIDS),
    DECLARE_PREFIX("AT+SUUID", AT_RESULT_SUUID),

    DECLARE_PREFIX("AT+WUUIDS", AT_RESULT_WUUIDS),
    DECLARE_PREFIX("AT+WUUID", AT_RESULT_WUUID),

    DECLARE_PREFIX("AT+NUUIDS", AT_RESULT_NUUIDS),
    DECLARE_PREFIX("AT+NUUID", AT_RESULT_NUUID),

    DECLARE_PREFIX("AT+ADVIS1\r\n", AT_RESULT_ADVIS1),
    DECLARE_PREFIX("AT+ADVIS2\r\n", AT_RESULT_ADVIS2),
    DECLARE_PREFIX("AT+ADVIS3\r\n", AT_RESULT_ADVIS3),
    DECLARE_PREFIX("AT+ADVIS4\r\n", AT_RESULT_ADVIS4),
    DECLARE_PREFIX("AT+ADVIS5\r\n", AT_RESULT_ADVIS5),
    DECLARE_PREFIX("AT+ADVI1\r\n", AT_RESULT_ADVI1),
    DECLARE_PREFIX("AT+ADVI2\r\n", AT_RESULT_ADVI2),
    DECLARE_PREFIX("AT+ADVI3\r\n", AT_RESULT_ADVI3),
    DECLARE_PREFIX("AT+ADVI4\r\n", AT_RESULT_ADVI4),
    DECLARE_PREFIX("AT+ADVI5\r\n", AT_RESULT_ADVI5),

    DECLARE_PREFIX("AT+RFLASH", AT_RESULT_RFLASH),
    DECLARE_PREFIX("AT+WFLASH", AT_RESULT_WFLASH),

    DECLARE_PREFIX("AT+TPMODE0\r\n", AT_RESULT_TPMODE0),
    DECLARE_PREFIX("AT+TPMODE1\r\n", AT_RESULT_TPMODE1),
};

static const at_prefix_t at_prefix_rsp[] = {
    DECLARE_PREFIX("OK", AT_RESULT_OK0),
    DECLARE_PREFIX("OK9600", AT_RESULT_OK1),
    DECLARE_PREFIX("OK19200", AT_RESULT_OK2),
    DECLARE_PREFIX("OK38400", AT_RESULT_OK3),
    DECLARE_PREFIX("OK57600", AT_RESULT_OK4),
    DECLARE_PREFIX("OK115200", AT_RESULT_OK5),

    DECLARE_PREFIX("OK100", AT_RESULT_ADVOK1),
    DECLARE_PREFIX("OK250", AT_RESULT_ADVOK2),
    DECLARE_PREFIX("OK500", AT_RESULT_ADVOK3),
    DECLARE_PREFIX("OK1600", AT_RESULT_ADVOK4),
    DECLARE_PREFIX("OK3200", AT_RESULT_ADVOK5),
    DECLARE_PREFIX("OK", AT_RESULT_OKDATA),
  
    DECLARE_PREFIX("OKTPMODE0", AT_RESULT_OKMODE0),
    DECLARE_PREFIX("OKTPMODE1", AT_RESULT_OKMODE1),
};

enum TAG_t
{
    TAG_ADV = 0xA0,
    TAG_BAUD,
    TAG_NAME,
    TAG_FFF0,
    TAG_FFF1,
    TAG_FFF2,
    TAG_MAX
};

#define extMCU_USED_FLASH_AREA 0x8000 // area no = 2
#define BK3432_USED_FLASH_AREA 0x80   // area no = 1,save device name,advertising interval parameter(1~5),

#define AT_CMD_IO_SET "AT+IOSET%1u\r\n"
#define AT_CMD_IO_GET "AT+IOGET\r\n"
typedef struct
{
    uint8_t *data;
    uint8_t len;
} at_rsp_content;

void save_parameter_to_BK3432_USED_FLASH_AREA(uint8_t TAG_value, uint8_t *buffer);
void load_parameter_from_BK3432_USED_FLASH_AREA(void);
at_rsp_content *at_wait_for_rsp(void);
/// @} APP

#endif //(BLE_APP_PRESENT)

#endif // APP_H_
