/**
 ****************************************************************************************
 *
 * @file app_f000.c
 *
 * @brief f000 Application Module entry point
 *
 * @auth  gang.cheng
 *
 * @date  2016.05.31
 *
 * Copyright (C) Beken 2009-2016
 *
 *
 ****************************************************************************************
 */

/**
 ****************************************************************************************
 * @addtogroup APP
 * @{
 ****************************************************************************************
 */

#include "rwip_config.h" // SW configuration

/*
 * INCLUDE FILES
 ****************************************************************************************
 */
#include <string.h>
#include "app.h"      // Application Definitions
#include "app_task.h" // application task definitions
#include "app_tp_mode.h"
#include "tp_mode_task.h" // health thermometer functions
#include "co_bt.h"
#include "prf_types.h" // Profile common types definition
#include "arch.h"      // Platform Definitions
#include "prf.h"
#include "tp_mode.h"
#include "ke_timer.h"
#include "uart.h"
#include "uart2.h"
#include "gpio.h"

/*
 * DEFINES
 ****************************************************************************************
 */

/*
 * GLOBAL VARIABLE DEFINITIONS
 ****************************************************************************************
 */

/// f000 Application Module Environment Structure
struct app_uart_tp_env_tag app_uart_tp_env;

/*
 * GLOBAL FUNCTION DEFINITIONS
 ****************************************************************************************
 */

void app_uart_tp_init(void)
{

	// Reset the environment
	memset(&app_uart_tp_env, 0, sizeof(struct app_uart_tp_env_tag));
}

void app_uart_add_uarts(void)
{

	struct uart_tp_db_cfg *db_cfg;

	struct gapm_profile_task_add_cmd *req = KE_MSG_ALLOC_DYN(GAPM_PROFILE_TASK_ADD_CMD,
			 TASK_GAPM, TASK_APP,
			 gapm_profile_task_add_cmd, sizeof(struct uart_tp_db_cfg));
	// Fill message
	req->operation = GAPM_PROFILE_TASK_ADD;
	req->sec_lvl = 0;
	req->prf_task_id = TASK_ID_UART_TP;
	req->app_task = TASK_APP;
	req->start_hdl = 0; // req->start_hdl = 0; dynamically allocated

	// Set parameters
	db_cfg = (struct uart_tp_db_cfg *)req->param;

	// Sending of notifications is supported
	db_cfg->features = UART_TP_TX_NTF_SUP;

	// Send the message
	ke_msg_send(req);
}

void app_uart_tx_send_payload(uint8_t *buf, uint8_t len)
{
	// Allocate the message
	struct uart_tx_payload_upd_req *req = KE_MSG_ALLOC(UART_TX_PAYLOAD_UPD_REQ,
		                       prf_get_task_from_id(TASK_ID_UART_TP),
		                       TASK_APP,
		                       uart_tx_payload_upd_req);
	// Fill in the parameter structure
	req->length = len;
	memcpy(req->tx_payload, buf, len);

	// Send the message
	ke_msg_send(req);
}

static int uart_tx_payload_ntf_cfg_ind_handler(ke_msg_id_t const msgid,
	                   struct uart_tx_payload_ntf_cfg_ind const *param,
	                   ke_task_id_t const dest_id,
	                   ke_task_id_t const src_id)
{
	UART_PRINTF("param->ntf_cfg = %x\r\n", param->ntf_cfg);
	if (param->ntf_cfg == PRF_CLI_STOP_NTFIND)
	{
		ke_timer_clear(UART_TX_PAYLOAD_PERIOD_NTF, dest_id);
	}
	else
	{
		ke_timer_set(UART_TX_PAYLOAD_PERIOD_NTF, dest_id, 1);
	}

	return (KE_MSG_CONSUMED);
}

static int uart_tx_payload_upd_handler(ke_msg_id_t const msgid,
	           struct uart_tx_payload_upd_rsp const *param,
	           ke_task_id_t const dest_id,
	           ke_task_id_t const src_id)
{
#if 0
  if(param->status == GAP_ERR_NO_ERROR)
  {
    uint8_t buf[128];
    memset(buf, 0xcc, 128);
    app_uart_tx_send_payload(buf, 128);
  }
#endif
	return (KE_MSG_CONSUMED);
}

/**
 ****************************************************************************************
 * @brief
 *
 * @param[in] msgid     Id of the message received.
 * @param[in] param     Pointer to the parameters of the message.
 * @param[in] dest_id   ID of the receiving task instance (TASK_GAP).
 * @param[in] src_id    ID of the sending task instance.
 *
 * @return If the message was consumed or not.
 ****************************************************************************************
 */
static int app_uart_msg_dflt_handler(ke_msg_id_t const msgid,
	         void const *param,
	         ke_task_id_t const dest_id,
	         ke_task_id_t const src_id)
{
	// Drop the message
	return (KE_MSG_CONSUMED);
}

static int uart_rx_writer_req_handler(ke_msg_id_t const msgid,
	          struct uart_rx_writer_ind *param,
	          ke_task_id_t const dest_id,
	          ke_task_id_t const src_id)
{
	// bypass payload to MCU
  uint8_t recv_data[256] = {0};
  memcpy(recv_data, param->uart_rx_payload, param->length);
  recv_data[param->length] = 0x0D;    //CR
  recv_data[param->length+1] = 0x0A;  //LF
	//uart_send((unsigned char *)param->uart_rx_payload, param->length);
  uart_send((unsigned char *)recv_data, param->length + 2);

	return (KE_MSG_CONSUMED);
}

static int uart_tx_period_ntf_handler(ke_msg_id_t const msgid,
	          struct uart_tx_payload_ntf_cfg_ind const *param,
	          ke_task_id_t const dest_id,
	          ke_task_id_t const src_id)
{
  #if 0
	uint8_t buf[128] = {0};
	memcpy(buf, "Notification Start\r\n", strlen("Notification Start\r\n"));
	app_uart_tx_send_payload(buf, 128);
  #endif
  
	return (KE_MSG_CONSUMED);
}

/*
 * LOCAL VARIABLE DEFINITIONS
 ****************************************************************************************
 */

/// Default State handlers definition
const struct ke_msg_handler app_uart_msg_handler_list[] =
    {
        // Note: first message is latest message checked by kernel so default is put on top.
        {KE_MSG_DEFAULT_HANDLER, (ke_msg_func_t)app_uart_msg_dflt_handler},
        {UART_TX_PAYLOAD_NTF_CFG_IND, (ke_msg_func_t)uart_tx_payload_ntf_cfg_ind_handler},
        {UART_TX_PAYLOAD_UPD_RSP, (ke_msg_func_t)uart_tx_payload_upd_handler},
        {UART_RX_WRITER_REQ_IND, (ke_msg_func_t)uart_rx_writer_req_handler},
        {UART_TX_PAYLOAD_PERIOD_NTF, (ke_msg_func_t)uart_tx_period_ntf_handler},
};

const struct ke_state_handler app_uart_table_handler =
    {&app_uart_msg_handler_list[0], (sizeof(app_uart_msg_handler_list) / sizeof(struct ke_msg_handler))};
